@@ -8,6 +8,7 @@
 //! Defines the AST structure of a graph DSL file
 
 use regex::Regex;
+use std::collections::HashMap;
 use std::fmt;
 use tree_sitter::CaptureQuantifier;
 use tree_sitter::Language;
@@ -20,22 +21,38 @@ use crate::Location;
 #[derive(Debug)]
 pub struct File {
     pub language: Language,
+    /// The expected global variables used in this file
+    pub globals: Vec<Global>,
     /// The combined query of all stanzas in the file
     pub query: Option<Query>,
     /// The list of stanzas in the file
     pub stanzas: Vec<Stanza>,
+    /// Attribute shorthands defined in the file
+    pub shorthands: AttributeShorthands,
 }
 
 impl File {
     pub fn new(language: Language) -> File {
         File {
             language,
+            globals: Vec::new(),
             query: None,
             stanzas: Vec::new(),
+            shorthands: AttributeShorthands::new(),
         }
     }
 }
 
+/// A global variable
+#[derive(Debug, Eq, PartialEq)]
+pub struct Global {
+    /// The name of the global variable
+    pub name: Identifier,
+    /// The quantifier of the global variable
+    pub quantifier: CaptureQuantifier,
+    pub location: Location,
+}
+
 /// One stanza within a file
 #[derive(Debug)]
 pub struct Stanza {
@@ -510,9 +527,12 @@ pub enum Expression {
     // Constants
     IntegerConstant(IntegerConstant),
     StringConstant(StringConstant),
+    // Literals
+    ListLiteral(ListLiteral),
+    SetLiteral(SetLiteral),
     // Comprehensions
-    List(ListComprehension),
-    Set(SetComprehension),
+    ListComprehension(ListComprehension),
+    SetComprehension(SetComprehension),
     // Syntax nodes
     Capture(Capture),
     // Variables
@@ -531,8 +551,10 @@ impl std::fmt::Display for Expression {
             Expression::TrueLiteral => write!(f, "true"),
             Expression::IntegerConstant(expr) => expr.fmt(f),
             Expression::StringConstant(expr) => expr.fmt(f),
-            Expression::List(expr) => expr.fmt(f),
-            Expression::Set(expr) => expr.fmt(f),
+            Expression::ListLiteral(expr) => expr.fmt(f),
+            Expression::SetLiteral(expr) => expr.fmt(f),
+            Expression::ListComprehension(expr) => expr.fmt(f),
+            Expression::SetComprehension(expr) => expr.fmt(f),
             Expression::Capture(expr) => expr.fmt(f),
             Expression::Variable(expr) => expr.fmt(f),
             Expression::Call(expr) => expr.fmt(f),
@@ -610,17 +632,17 @@ impl std::fmt::Display for IntegerConstant {
 
 /// An ordered list of values
 #[derive(Debug, Eq, PartialEq)]
-pub struct ListComprehension {
+pub struct ListLiteral {
     pub elements: Vec<Expression>,
 }
 
-impl From<ListComprehension> for Expression {
-    fn from(expr: ListComprehension) -> Expression {
-        Expression::List(expr)
+impl From<ListLiteral> for Expression {
+    fn from(expr: ListLiteral) -> Expression {
+        Expression::ListLiteral(expr)
     }
 }
 
-impl std::fmt::Display for ListComprehension {
+impl std::fmt::Display for ListLiteral {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "[")?;
         let mut first = true;
@@ -636,6 +658,31 @@ impl std::fmt::Display for ListComprehension {
     }
 }
 
+/// An list comprehension
+#[derive(Debug, Eq, PartialEq)]
+pub struct ListComprehension {
+    pub element: Box<Expression>,
+    pub variable: UnscopedVariable,
+    pub value: Box<Expression>,
+    pub location: Location,
+}
+
+impl From<ListComprehension> for Expression {
+    fn from(expr: ListComprehension) -> Expression {
+        Expression::ListComprehension(expr)
+    }
+}
+
+impl std::fmt::Display for ListComprehension {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(
+            f,
+            "[ {} for {} in {} ]",
+            self.element, self.variable, self.value
+        )
+    }
+}
+
 /// A reference to one of the regex captures in a `scan` statement
 #[derive(Debug, Eq, PartialEq)]
 pub struct RegexCapture {
@@ -656,17 +703,17 @@ impl std::fmt::Display for RegexCapture {
 
 /// An unordered set of values
 #[derive(Debug, Eq, PartialEq)]
-pub struct SetComprehension {
+pub struct SetLiteral {
     pub elements: Vec<Expression>,
 }
 
-impl From<SetComprehension> for Expression {
-    fn from(expr: SetComprehension) -> Expression {
-        Expression::Set(expr)
+impl From<SetLiteral> for Expression {
+    fn from(expr: SetLiteral) -> Expression {
+        Expression::SetLiteral(expr)
     }
 }
 
-impl std::fmt::Display for SetComprehension {
+impl std::fmt::Display for SetLiteral {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{{")?;
         let mut first = true;
@@ -682,6 +729,31 @@ impl std::fmt::Display for SetComprehension {
     }
 }
 
+/// An set comprehension
+#[derive(Debug, Eq, PartialEq)]
+pub struct SetComprehension {
+    pub element: Box<Expression>,
+    pub variable: UnscopedVariable,
+    pub value: Box<Expression>,
+    pub location: Location,
+}
+
+impl From<SetComprehension> for Expression {
+    fn from(expr: SetComprehension) -> Expression {
+        Expression::SetComprehension(expr)
+    }
+}
+
+impl std::fmt::Display for SetComprehension {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(
+            f,
+            "{{ {} for {} in {} }}",
+            self.element, self.variable, self.value
+        )
+    }
+}
+
 /// A string constant
 #[derive(Debug, Eq, PartialEq)]
 pub struct StringConstant {
@@ -717,3 +789,48 @@ impl From<ScopedVariable> for Expression {
         Expression::Variable(variable.into())
     }
 }
+
+/// Attribute shorthands
+#[derive(Debug, Eq, PartialEq)]
+pub struct AttributeShorthands(HashMap<Identifier, AttributeShorthand>);
+
+impl AttributeShorthands {
+    pub fn new() -> Self {
+        Self(HashMap::new())
+    }
+
+    pub fn get(&self, name: &Identifier) -> Option<&AttributeShorthand> {
+        self.0.get(name)
+    }
+
+    pub fn add(&mut self, shorthand: AttributeShorthand) {
+        self.0.insert(shorthand.name.clone(), shorthand);
+    }
+
+    pub fn iter(&self) -> impl Iterator<Item = &AttributeShorthand> {
+        self.0.values()
+    }
+
+    pub fn into_iter(self) -> impl Iterator<Item = AttributeShorthand> {
+        self.0.into_values()
+    }
+}
+
+/// An attribute shorthand
+#[derive(Debug, Eq, PartialEq)]
+pub struct AttributeShorthand {
+    pub name: Identifier,
+    pub variable: UnscopedVariable,
+    pub attributes: Vec<Attribute>,
+    pub location: Location,
+}
+
+impl std::fmt::Display for AttributeShorthand {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "attribute {} = {} =>", self.name, self.variable,)?;
+        for attr in &self.attributes {
+            write!(f, " {}", attr)?;
+        }
+        write!(f, " at {}", self.location)
+    }
+}
