@@ -6,6 +6,7 @@
 // ------------------------------------------------------------------------------------------------
 
 use anyhow::Context as ErrorContext;
+use std::collections::BTreeSet;
 use std::collections::HashMap;
 use thiserror::Error;
 use tree_sitter::CaptureQuantifier;
@@ -16,6 +17,9 @@ use tree_sitter::Tree;
 use crate::ast::AddEdgeAttribute;
 use crate::ast::AddGraphNodeAttribute;
 use crate::ast::Assign;
+use crate::ast::Attribute;
+use crate::ast::AttributeShorthand;
+use crate::ast::AttributeShorthands;
 use crate::ast::Call;
 use crate::ast::Capture;
 use crate::ast::Condition;
@@ -29,23 +33,27 @@ use crate::ast::ForIn;
 use crate::ast::If;
 use crate::ast::IntegerConstant;
 use crate::ast::ListComprehension;
+use crate::ast::ListLiteral;
 use crate::ast::Print;
 use crate::ast::RegexCapture;
 use crate::ast::Scan;
 use crate::ast::ScopedVariable;
 use crate::ast::SetComprehension;
+use crate::ast::SetLiteral;
 use crate::ast::Stanza;
 use crate::ast::Statement;
 use crate::ast::StringConstant;
 use crate::ast::UnscopedVariable;
 use crate::ast::Variable;
 use crate::functions::Functions;
+use crate::graph::Attributes;
 use crate::graph::Graph;
 use crate::graph::SyntaxNodeRef;
 use crate::graph::Value;
-pub use crate::variables::Globals;
+use crate::variables::Globals;
 use crate::variables::VariableMap;
 use crate::variables::Variables;
+use crate::Identifier;
 
 impl File {
     /// Executes this graph DSL file against a source file.  You must provide the parsed syntax
@@ -55,11 +63,10 @@ impl File {
         &self,
         tree: &'tree Tree,
         source: &'tree str,
-        functions: &mut Functions,
-        globals: &Globals,
+        config: &mut ExecutionConfig,
     ) -> Result<Graph<'tree>, ExecutionError> {
         let mut graph = Graph::new();
-        self.execute_into(&mut graph, tree, source, functions, globals)?;
+        self.execute_into(&mut graph, tree, source, config)?;
         Ok(graph)
     }
 
@@ -73,12 +80,30 @@ impl File {
         graph: &mut Graph<'tree>,
         tree: &'tree Tree,
         source: &'tree str,
-        functions: &mut Functions,
-        globals: &Globals,
+        config: &mut ExecutionConfig,
     ) -> Result<(), ExecutionError> {
-        if tree.root_node().has_error() {
-            return Err(ExecutionError::ParseTreeHasErrors);
+        if config.lazy {
+            self.execute_lazy_into(graph, tree, source, config)
+        } else {
+            self.execute_strict_into(graph, tree, source, config)
         }
+    }
+}
+
+impl File {
+    /// Executes this graph DSL file against a source file, saving the results into an existing
+    /// `Graph` instance.  You must provide the parsed syntax tree (`tree`) as well as the source
+    /// text that it was parsed from (`source`).  You also provide the set of functions and global
+    /// variables that are available during execution. This variant is useful when you need to
+    /// “pre-seed” the graph with some predefined nodes and/or edges before executing the DSL file.
+    fn execute_strict_into<'tree>(
+        &self,
+        graph: &mut Graph<'tree>,
+        tree: &'tree Tree,
+        source: &'tree str,
+        config: &mut ExecutionConfig,
+    ) -> Result<(), ExecutionError> {
+        self.check_globals(config.globals)?;
         let mut locals = VariableMap::new();
         let mut scoped = ScopedVariables::new();
         let current_regex_captures = Vec::new();
@@ -89,17 +114,86 @@ impl File {
                 tree,
                 source,
                 graph,
-                functions,
-                globals,
+                config,
                 &mut locals,
                 &mut scoped,
                 &current_regex_captures,
                 &mut function_parameters,
                 &mut cursor,
+                &self.shorthands,
             )?;
         }
         Ok(())
     }
+
+    pub(crate) fn check_globals(&self, globals: &Globals) -> Result<(), ExecutionError> {
+        for global in &self.globals {
+            match globals.get(&global.name) {
+                None => {
+                    return Err(ExecutionError::MissingGlobalVariable(
+                        global.name.as_str().to_string(),
+                    ));
+                }
+                Some(value) => {
+                    if global.quantifier == CaptureQuantifier::ZeroOrMore
+                        || global.quantifier == CaptureQuantifier::OneOrMore
+                    {
+                        if value.as_list().is_err() {
+                            return Err(ExecutionError::ExpectedList(
+                                global.name.as_str().to_string(),
+                            ));
+                        }
+                    }
+                }
+            }
+        }
+        Ok(())
+    }
+}
+
+/// Configuration for the execution of a File
+pub struct ExecutionConfig<'a, 'g> {
+    pub(crate) functions: &'a mut Functions,
+    pub(crate) globals: &'a Globals<'g>,
+    pub(crate) lazy: bool,
+    pub(crate) location_attr: Option<Identifier>,
+    pub(crate) variable_name_attr: Option<Identifier>,
+}
+
+impl<'a, 'g> ExecutionConfig<'a, 'g> {
+    pub fn new(functions: &'a mut Functions, globals: &'a Globals<'g>) -> Self {
+        Self {
+            functions,
+            globals,
+            lazy: false,
+            location_attr: None,
+            variable_name_attr: None,
+        }
+    }
+
+    pub fn debug_attributes(
+        self,
+        location_attr: Identifier,
+        variable_name_attr: Identifier,
+    ) -> Self {
+        Self {
+            functions: self.functions,
+            globals: self.globals,
+            lazy: self.lazy,
+            location_attr: location_attr.into(),
+            variable_name_attr: variable_name_attr.into(),
+        }
+    }
+
+    pub fn lazy(self, lazy: bool) -> Self {
+        Self {
+            functions: self.functions,
+            globals: self.globals,
+            lazy,
+            location_attr: self.location_attr,
+            variable_name_attr: self.variable_name_attr,
+        }
+    }
 }
 
 /// An error that can occur while executing a graph DSL file
@@ -133,6 +227,8 @@ pub enum ExecutionError {
     InvalidParameters(String),
     #[error("Scoped variables can only be attached to syntax nodes {0}")]
     InvalidVariableScope(String),
+    #[error("Missing global variable {0}")]
+    MissingGlobalVariable(String),
     #[error("Recursively defined scoped variable {0}")]
     RecursivelyDefinedScopedVariable(String),
     #[error("Recursively defined variable {0}")]
@@ -153,8 +249,6 @@ pub enum ExecutionError {
     UndefinedVariable(String),
     #[error("Cannot add scoped variable after being forced {0}")]
     VariableScopesAlreadyForced(String),
-    #[error("Parse tree has errors")]
-    ParseTreeHasErrors,
     #[error(transparent)]
     Other(#[from] anyhow::Error),
 }
@@ -169,16 +263,16 @@ impl ExecutionError {
     }
 }
 
-struct ExecutionContext<'a, 'g, 's, 'tree> {
+struct ExecutionContext<'a, 'c, 'g, 's, 'tree> {
     source: &'tree str,
     graph: &'a mut Graph<'tree>,
-    functions: &'a mut Functions,
-    globals: &'a Globals<'g>,
+    config: &'a mut ExecutionConfig<'c, 'g>,
     locals: &'a mut dyn Variables<Value>,
     scoped: &'a mut ScopedVariables<'s>,
     current_regex_captures: &'a Vec<String>,
     function_parameters: &'a mut Vec<Value>,
     mat: &'a QueryMatch<'a, 'tree>,
+    shorthands: &'a AttributeShorthands,
 }
 
 struct ScopedVariables<'a> {
@@ -198,18 +292,18 @@ impl<'a> ScopedVariables<'a> {
 }
 
 impl Stanza {
-    fn execute<'l, 's, 'tree>(
+    fn execute<'g, 'l, 's, 'tree>(
         &self,
         tree: &'tree Tree,
         source: &'tree str,
         graph: &mut Graph<'tree>,
-        functions: &mut Functions,
-        globals: &Globals,
+        config: &mut ExecutionConfig<'_, 'g>,
         locals: &mut VariableMap<'l, Value>,
         scoped: &mut ScopedVariables<'s>,
         current_regex_captures: &Vec<String>,
         function_parameters: &mut Vec<Value>,
         cursor: &mut QueryCursor,
+        shorthands: &AttributeShorthands,
     ) -> Result<(), ExecutionError> {
         let matches = cursor.matches(&self.query, tree.root_node(), source.as_bytes());
         for mat in matches {
@@ -217,13 +311,13 @@ impl Stanza {
             let mut exec = ExecutionContext {
                 source,
                 graph,
-                functions,
-                globals,
+                config,
                 locals,
                 scoped,
                 current_regex_captures,
                 function_parameters,
                 mat: &mat,
+                shorthands: shorthands,
             };
             for statement in &self.statements {
                 statement
@@ -277,6 +371,8 @@ impl Assign {
 impl CreateGraphNode {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let graph_node = exec.graph.add_graph_node();
+        self.node
+            .add_debug_attrs(&mut exec.graph[graph_node].attributes, exec.config)?;
         let value = Value::GraphNode(graph_node);
         self.node.add(exec, value, false)
     }
@@ -285,17 +381,19 @@ impl CreateGraphNode {
 impl AddGraphNodeAttribute {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let node = self.node.evaluate(exec)?.into_graph_node_ref()?;
-        for attribute in &self.attributes {
-            let value = attribute.value.evaluate(exec)?;
+        let add_attribute = |exec: &mut ExecutionContext, name: Identifier, value: Value| {
             exec.graph[node]
                 .attributes
-                .add(attribute.name.clone(), value)
+                .add(name.clone(), value)
                 .map_err(|_| {
                     ExecutionError::DuplicateAttribute(format!(
                         " {} on graph node ({}) in {}",
-                        attribute.name, node, self,
+                        name, node, self,
                     ))
-                })?;
+                })
+        };
+        for attribute in &self.attributes {
+            attribute.execute(exec, &add_attribute)?;
         }
         Ok(())
     }
@@ -319,8 +417,7 @@ impl AddEdgeAttribute {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let source = self.source.evaluate(exec)?.into_graph_node_ref()?;
         let sink = self.sink.evaluate(exec)?.into_graph_node_ref()?;
-        for attribute in &self.attributes {
-            let value = attribute.value.evaluate(exec)?;
+        let add_attribute = |exec: &mut ExecutionContext, name: Identifier, value: Value| {
             let edge = match exec.graph[source].get_edge_mut(sink) {
                 Some(edge) => Ok(edge),
                 None => Err(ExecutionError::UndefinedEdge(format!(
@@ -328,14 +425,15 @@ impl AddEdgeAttribute {
                     source, sink, self,
                 ))),
             }?;
-            edge.attributes
-                .add(attribute.name.clone(), value)
-                .map_err(|_| {
-                    ExecutionError::DuplicateAttribute(format!(
-                        " {} on edge ({} -> {}) in {}",
-                        attribute.name, source, sink, self,
-                    ))
-                })?;
+            edge.attributes.add(name.clone(), value).map_err(|_| {
+                ExecutionError::DuplicateAttribute(format!(
+                    " {} on edge ({} -> {}) in {}",
+                    name, source, sink, self,
+                ))
+            })
+        };
+        for attribute in &self.attributes {
+            attribute.execute(exec, &add_attribute)?;
         }
         Ok(())
     }
@@ -380,17 +478,17 @@ impl Scan {
                     .push(regex_capture.map(|m| m.as_str()).unwrap_or("").to_string());
             }
 
-            let mut arm_locals = VariableMap::new_child(exec.locals);
+            let mut arm_locals = VariableMap::nested(exec.locals);
             let mut arm_exec = ExecutionContext {
                 source: exec.source,
                 graph: exec.graph,
-                functions: exec.functions,
-                globals: exec.globals,
+                config: exec.config,
                 locals: &mut arm_locals,
                 scoped: exec.scoped,
                 current_regex_captures: &current_regex_captures,
                 function_parameters: exec.function_parameters,
                 mat: exec.mat,
+                shorthands: exec.shorthands,
             };
 
             for statement in &arm.statements {
@@ -435,17 +533,17 @@ impl If {
                 result &= condition.test(exec)?;
             }
             if result {
-                let mut arm_locals = VariableMap::new_child(exec.locals);
+                let mut arm_locals = VariableMap::nested(exec.locals);
                 let mut arm_exec = ExecutionContext {
                     source: exec.source,
                     graph: exec.graph,
-                    functions: exec.functions,
-                    globals: exec.globals,
+                    config: exec.config,
                     locals: &mut arm_locals,
                     scoped: exec.scoped,
                     current_regex_captures: exec.current_regex_captures,
                     function_parameters: exec.function_parameters,
                     mat: exec.mat,
+                    shorthands: exec.shorthands,
                 };
                 for stmt in &arm.statements {
                     stmt.execute(&mut arm_exec)?;
@@ -470,19 +568,19 @@ impl Condition {
 impl ForIn {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let values = self.value.evaluate(exec)?.into_list()?;
-        let mut loop_locals = VariableMap::new_child(exec.locals);
+        let mut loop_locals = VariableMap::nested(exec.locals);
         for value in values {
             loop_locals.clear();
             let mut loop_exec = ExecutionContext {
                 source: exec.source,
                 graph: exec.graph,
-                functions: exec.functions,
-                globals: exec.globals,
+                config: exec.config,
                 locals: &mut loop_locals,
                 scoped: exec.scoped,
                 current_regex_captures: exec.current_regex_captures,
                 function_parameters: exec.function_parameters,
                 mat: exec.mat,
+                shorthands: exec.shorthands,
             };
             self.variable.add(&mut loop_exec, value, false)?;
             for stmt in &self.statements {
@@ -501,8 +599,10 @@ impl Expression {
             Expression::TrueLiteral => Ok(Value::Boolean(true)),
             Expression::IntegerConstant(expr) => expr.evaluate(exec),
             Expression::StringConstant(expr) => expr.evaluate(exec),
-            Expression::List(expr) => expr.evaluate(exec),
-            Expression::Set(expr) => expr.evaluate(exec),
+            Expression::ListLiteral(expr) => expr.evaluate(exec),
+            Expression::SetLiteral(expr) => expr.evaluate(exec),
+            Expression::ListComprehension(expr) => expr.evaluate(exec),
+            Expression::SetComprehension(expr) => expr.evaluate(exec),
             Expression::Capture(expr) => expr.evaluate(exec),
             Expression::Variable(expr) => expr.evaluate(exec),
             Expression::Call(expr) => expr.evaluate(exec),
@@ -523,7 +623,7 @@ impl StringConstant {
     }
 }
 
-impl ListComprehension {
+impl ListLiteral {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
         let elements = self
             .elements
@@ -534,7 +634,33 @@ impl ListComprehension {
     }
 }
 
-impl SetComprehension {
+impl ListComprehension {
+    fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
+        let values = self.value.evaluate(exec)?.into_list()?;
+        let mut elements = Vec::new();
+        let mut loop_locals = VariableMap::nested(exec.locals);
+        for value in values {
+            loop_locals.clear();
+            let mut loop_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                config: exec.config,
+                locals: &mut loop_locals,
+                scoped: exec.scoped,
+                current_regex_captures: exec.current_regex_captures,
+                function_parameters: exec.function_parameters,
+                mat: exec.mat,
+                shorthands: exec.shorthands,
+            };
+            self.variable.add(&mut loop_exec, value, false)?;
+            let element = self.element.evaluate(&mut loop_exec)?;
+            elements.push(element);
+        }
+        Ok(Value::List(elements))
+    }
+}
+
+impl SetLiteral {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
         let elements = self
             .elements
@@ -545,6 +671,32 @@ impl SetComprehension {
     }
 }
 
+impl SetComprehension {
+    fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
+        let values = self.value.evaluate(exec)?.into_list()?;
+        let mut elements = BTreeSet::new();
+        let mut loop_locals = VariableMap::nested(exec.locals);
+        for value in values {
+            loop_locals.clear();
+            let mut loop_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                config: exec.config,
+                locals: &mut loop_locals,
+                scoped: exec.scoped,
+                current_regex_captures: exec.current_regex_captures,
+                function_parameters: exec.function_parameters,
+                mat: exec.mat,
+                shorthands: exec.shorthands,
+            };
+            self.variable.add(&mut loop_exec, value, false)?;
+            let element = self.element.evaluate(&mut loop_exec)?;
+            elements.insert(element);
+        }
+        Ok(Value::Set(elements))
+    }
+}
+
 impl Capture {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
         Ok(query_capture_value(
@@ -562,7 +714,7 @@ impl Call {
             let parameter = parameter.evaluate(exec)?;
             exec.function_parameters.push(parameter);
         }
-        exec.functions.call(
+        exec.config.functions.call(
             &self.function,
             exec.graph,
             exec.source,
@@ -683,7 +835,7 @@ impl ScopedVariable {
 
 impl UnscopedVariable {
     fn get<'a>(&self, exec: &'a mut ExecutionContext) -> Result<&'a Value, ExecutionError> {
-        if let Some(value) = exec.globals.get(&self.name) {
+        if let Some(value) = exec.config.globals.get(&self.name) {
             Some(value)
         } else {
             exec.locals.get(&self.name)
@@ -697,7 +849,7 @@ impl UnscopedVariable {
         value: Value,
         mutable: bool,
     ) -> Result<(), ExecutionError> {
-        if exec.globals.get(&self.name).is_some() {
+        if exec.config.globals.get(&self.name).is_some() {
             return Err(ExecutionError::DuplicateVariable(format!(
                 " global {}",
                 self,
@@ -709,7 +861,7 @@ impl UnscopedVariable {
     }
 
     fn set(&self, exec: &mut ExecutionContext, value: Value) -> Result<(), ExecutionError> {
-        if exec.globals.get(&self.name).is_some() {
+        if exec.config.globals.get(&self.name).is_some() {
             return Err(ExecutionError::CannotAssignImmutableVariable(format!(
                 " global {}",
                 self,
@@ -725,6 +877,80 @@ impl UnscopedVariable {
     }
 }
 
+impl Variable {
+    pub(crate) fn add_debug_attrs(
+        &self,
+        attributes: &mut Attributes,
+        config: &ExecutionConfig,
+    ) -> Result<(), ExecutionError> {
+        if let Some(variable_name_attr) = &config.variable_name_attr {
+            attributes
+                .add(variable_name_attr.clone(), format!("{}", self))
+                .map_err(|_| {
+                    ExecutionError::DuplicateAttribute(variable_name_attr.as_str().into())
+                })?;
+        }
+        if let Some(location_attr) = &config.location_attr {
+            let location = match &self {
+                Variable::Scoped(v) => v.location,
+                Variable::Unscoped(v) => v.location,
+            };
+            attributes
+                .add(location_attr.clone(), format!("{}", location))
+                .map_err(|_| ExecutionError::DuplicateAttribute(location_attr.as_str().into()))?;
+        }
+        Ok(())
+    }
+}
+
+impl Attribute {
+    fn execute<F>(
+        &self,
+        exec: &mut ExecutionContext,
+        add_attribute: &F,
+    ) -> Result<(), ExecutionError>
+    where
+        F: Fn(&mut ExecutionContext, Identifier, Value) -> Result<(), ExecutionError>,
+    {
+        let value = self.value.evaluate(exec)?;
+        if let Some(shorthand) = exec.shorthands.get(&self.name) {
+            shorthand.execute(exec, add_attribute, value)
+        } else {
+            add_attribute(exec, self.name.clone(), value)
+        }
+    }
+}
+
+impl AttributeShorthand {
+    fn execute<F>(
+        &self,
+        exec: &mut ExecutionContext,
+        add_attribute: &F,
+        value: Value,
+    ) -> Result<(), ExecutionError>
+    where
+        F: Fn(&mut ExecutionContext, Identifier, Value) -> Result<(), ExecutionError>,
+    {
+        let mut shorthand_locals = VariableMap::new();
+        let mut shorthand_exec = ExecutionContext {
+            source: exec.source,
+            graph: exec.graph,
+            config: exec.config,
+            locals: &mut shorthand_locals,
+            scoped: exec.scoped,
+            current_regex_captures: exec.current_regex_captures,
+            function_parameters: exec.function_parameters,
+            mat: exec.mat,
+            shorthands: exec.shorthands,
+        };
+        self.variable.add(&mut shorthand_exec, value, false)?;
+        for attr in &self.attributes {
+            attr.execute(&mut shorthand_exec, add_attribute)?;
+        }
+        Ok(())
+    }
+}
+
 /// Get the value for the given capture, considering the suffix
 pub(crate) fn query_capture_value<'tree>(
     index: usize,
