@@ -21,11 +21,11 @@ use tree_sitter::Tree;
 
 use crate::ast;
 use crate::execution::query_capture_value;
+use crate::execution::ExecutionConfig;
 use crate::execution::ExecutionError;
 use crate::functions::Functions;
 use crate::graph;
 use crate::graph::Graph;
-use crate::variables::Globals;
 use crate::variables::VariableMap;
 use crate::variables::Variables;
 use crate::Identifier;
@@ -35,37 +35,19 @@ use store::*;
 use values::*;
 
 impl ast::File {
-    /// Executes this graph DSL file against a source file.  You must provide the parsed syntax
-    /// tree (`tree`) as well as the source text that it was parsed from (`source`).  You also
-    /// provide the set of functions and global variables that are available during execution.
-    pub fn execute_lazy<'tree>(
-        &self,
-        tree: &'tree Tree,
-        source: &'tree str,
-        functions: &mut Functions,
-        globals: &Globals,
-    ) -> Result<Graph<'tree>, ExecutionError> {
-        let mut graph = Graph::new();
-        self.execute_lazy_into(&mut graph, tree, source, functions, globals)?;
-        Ok(graph)
-    }
-
     /// Executes this graph DSL file against a source file, saving the results into an existing
     /// `Graph` instance.  You must provide the parsed syntax tree (`tree`) as well as the source
     /// text that it was parsed from (`source`).  You also provide the set of functions and global
     /// variables that are available during execution. This variant is useful when you need to
     /// “pre-seed” the graph with some predefined nodes and/or edges before executing the DSL file.
-    pub fn execute_lazy_into<'tree>(
+    pub(crate) fn execute_lazy_into<'tree>(
         &self,
         graph: &mut Graph<'tree>,
         tree: &'tree Tree,
         source: &'tree str,
-        functions: &mut Functions,
-        globals: &Globals,
+        config: &mut ExecutionConfig,
     ) -> Result<(), ExecutionError> {
-        if tree.root_node().has_error() {
-            return Err(ExecutionError::ParseTreeHasErrors);
-        }
+        self.check_globals(config.globals)?;
         let mut locals = VariableMap::new();
         let mut cursor = QueryCursor::new();
         let mut store = LazyStore::new();
@@ -82,14 +64,14 @@ impl ast::File {
                 source,
                 &mat,
                 graph,
-                functions,
-                globals,
+                config,
                 &mut locals,
                 &mut store,
                 &mut scoped_store,
                 &mut lazy_graph,
                 &mut function_parameters,
                 &mut prev_element_debug_info,
+                &self.shorthands,
             )?;
         }
 
@@ -98,7 +80,7 @@ impl ast::File {
                 .evaluate(&mut EvaluationContext {
                     source,
                     graph,
-                    functions,
+                    functions: config.functions,
                     store: &mut store,
                     scoped_store: &mut scoped_store,
                     function_parameters: &mut function_parameters,
@@ -112,11 +94,10 @@ impl ast::File {
 }
 
 /// Context for execution, which executes stanzas to build the lazy graph
-struct ExecutionContext<'a, 'g, 'tree> {
+struct ExecutionContext<'a, 'c, 'g, 'tree> {
     source: &'tree str,
     graph: &'a mut Graph<'tree>,
-    functions: &'a mut Functions,
-    globals: &'a Globals<'g>,
+    config: &'a mut ExecutionConfig<'c, 'g>,
     locals: &'a mut dyn Variables<LazyValue>,
     current_regex_captures: &'a Vec<String>,
     mat: &'a QueryMatch<'a, 'tree>,
@@ -125,6 +106,7 @@ struct ExecutionContext<'a, 'g, 'tree> {
     lazy_graph: &'a mut Vec<LazyStatement>,
     function_parameters: &'a mut Vec<graph::Value>, // re-usable buffer to reduce memory allocations
     prev_element_debug_info: &'a mut HashMap<GraphElementKey, DebugInfo>,
+    shorthands: &'a ast::AttributeShorthands,
 }
 
 /// Context for evaluation, which evalautes the lazy graph to build the actual graph
@@ -146,27 +128,26 @@ pub(super) enum GraphElementKey {
 }
 
 impl ast::Stanza {
-    fn execute_lazy<'l, 'g, 'q, 'tree>(
+    fn execute_lazy<'a, 'l, 'g, 'q, 'tree>(
         &self,
         source: &'tree str,
         mat: &QueryMatch<'_, 'tree>,
         graph: &mut Graph<'tree>,
-        functions: &mut Functions,
-        globals: &Globals<'g>,
+        config: &mut ExecutionConfig,
         locals: &mut VariableMap<'l, LazyValue>,
         store: &mut LazyStore,
         scoped_store: &mut LazyScopedVariables,
         lazy_graph: &mut Vec<LazyStatement>,
         function_parameters: &mut Vec<graph::Value>,
         prev_element_debug_info: &mut HashMap<GraphElementKey, DebugInfo>,
+        shorthands: &ast::AttributeShorthands,
     ) -> Result<(), ExecutionError> {
         let current_regex_captures = vec![];
         locals.clear();
         let mut exec = ExecutionContext {
             source,
             graph,
-            functions,
-            globals,
+            config,
             locals,
             current_regex_captures: &current_regex_captures,
             mat,
@@ -175,6 +156,7 @@ impl ast::Stanza {
             lazy_graph,
             function_parameters,
             prev_element_debug_info,
+            shorthands,
         };
         let node = query_capture_value(self.full_match_file_capture_index, One, &mat, exec.graph);
         debug!("match {} at {}", node, self.location);
@@ -231,6 +213,8 @@ impl ast::Assign {
 impl ast::CreateGraphNode {
     fn execute_lazy(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let graph_node = exec.graph.add_graph_node();
+        self.node
+            .add_debug_attrs(&mut exec.graph[graph_node].attributes, exec.config)?;
         self.node.add_lazy(exec, graph_node.into(), false)
     }
 }
@@ -239,8 +223,9 @@ impl ast::AddGraphNodeAttribute {
     fn execute_lazy(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let node = self.node.evaluate_lazy(exec)?;
         let mut attributes = Vec::new();
+        let mut add_attribute = |a| attributes.push(a);
         for attribute in &self.attributes {
-            attributes.push(attribute.evaluate_lazy(exec)?);
+            attribute.execute_lazy(exec, &mut add_attribute)?;
         }
         let stmt = LazyAddGraphNodeAttribute::new(node, attributes, self.location.into());
         exec.lazy_graph.push(stmt.into());
@@ -263,8 +248,9 @@ impl ast::AddEdgeAttribute {
         let source = self.source.evaluate_lazy(exec)?;
         let sink = self.sink.evaluate_lazy(exec)?;
         let mut attributes = Vec::new();
+        let mut add_attribute = |a| attributes.push(a);
         for attribute in &self.attributes {
-            attributes.push(attribute.evaluate_lazy(exec)?);
+            attribute.execute_lazy(exec, &mut add_attribute)?;
         }
         let stmt = LazyAddEdgeAttribute::new(source, sink, attributes, self.location.into());
         exec.lazy_graph.push(stmt.into());
@@ -311,12 +297,11 @@ impl ast::Scan {
                     .push(regex_capture.map(|m| m.as_str()).unwrap_or("").to_string());
             }
 
-            let mut arm_locals = VariableMap::new_child(exec.locals);
+            let mut arm_locals = VariableMap::nested(exec.locals);
             let mut arm_exec = ExecutionContext {
                 source: exec.source,
                 graph: exec.graph,
-                functions: exec.functions,
-                globals: exec.globals,
+                config: exec.config,
                 locals: &mut arm_locals,
                 current_regex_captures: &current_regex_captures,
                 mat: exec.mat,
@@ -325,6 +310,7 @@ impl ast::Scan {
                 lazy_graph: exec.lazy_graph,
                 function_parameters: exec.function_parameters,
                 prev_element_debug_info: exec.prev_element_debug_info,
+                shorthands: exec.shorthands,
             };
 
             for statement in &arm.statements {
@@ -371,12 +357,11 @@ impl ast::If {
                 result &= condition.test_eager(exec)?;
             }
             if result {
-                let mut arm_locals = VariableMap::new_child(exec.locals);
+                let mut arm_locals = VariableMap::nested(exec.locals);
                 let mut arm_exec = ExecutionContext {
                     source: exec.source,
                     graph: exec.graph,
-                    functions: exec.functions,
-                    globals: exec.globals,
+                    config: exec.config,
                     locals: &mut arm_locals,
                     current_regex_captures: exec.current_regex_captures,
                     mat: exec.mat,
@@ -385,6 +370,7 @@ impl ast::If {
                     lazy_graph: exec.lazy_graph,
                     function_parameters: exec.function_parameters,
                     prev_element_debug_info: exec.prev_element_debug_info,
+                    shorthands: exec.shorthands,
                 };
                 for stmt in &arm.statements {
                     stmt.execute_lazy(&mut arm_exec)?;
@@ -411,14 +397,13 @@ impl ast::Condition {
 impl ast::ForIn {
     fn execute_lazy(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let values = self.value.evaluate_eager(exec)?.into_list()?;
-        let mut loop_locals = VariableMap::new_child(exec.locals);
+        let mut loop_locals = VariableMap::nested(exec.locals);
         for value in values {
             loop_locals.clear();
             let mut loop_exec = ExecutionContext {
                 source: exec.source,
                 graph: exec.graph,
-                functions: exec.functions,
-                globals: exec.globals,
+                config: exec.config,
                 locals: &mut loop_locals,
                 current_regex_captures: exec.current_regex_captures,
                 mat: exec.mat,
@@ -427,6 +412,7 @@ impl ast::ForIn {
                 lazy_graph: exec.lazy_graph,
                 function_parameters: exec.function_parameters,
                 prev_element_debug_info: exec.prev_element_debug_info,
+                shorthands: exec.shorthands,
             };
             self.variable
                 .add_lazy(&mut loop_exec, value.into(), false)?;
@@ -446,8 +432,10 @@ impl ast::Expression {
             Self::TrueLiteral => Ok(true.into()),
             Self::IntegerConstant(expr) => expr.evaluate_lazy(exec),
             Self::StringConstant(expr) => expr.evaluate_lazy(exec),
-            Self::List(expr) => expr.evaluate_lazy(exec),
-            Self::Set(expr) => expr.evaluate_lazy(exec),
+            Self::ListLiteral(expr) => expr.evaluate_lazy(exec),
+            Self::SetLiteral(expr) => expr.evaluate_lazy(exec),
+            Self::ListComprehension(expr) => expr.evaluate_lazy(exec),
+            Self::SetComprehension(expr) => expr.evaluate_lazy(exec),
             Self::Capture(expr) => expr.evaluate_lazy(exec),
             Self::Variable(expr) => expr.evaluate_lazy(exec),
             Self::Call(expr) => expr.evaluate_lazy(exec),
@@ -461,7 +449,7 @@ impl ast::Expression {
         self.evaluate_lazy(exec)?.evaluate(&mut EvaluationContext {
             source: exec.source,
             graph: exec.graph,
-            functions: exec.functions,
+            functions: exec.config.functions,
             store: exec.store,
             scoped_store: exec.scoped_store,
             function_parameters: exec.function_parameters,
@@ -482,7 +470,7 @@ impl ast::StringConstant {
     }
 }
 
-impl ast::ListComprehension {
+impl ast::ListLiteral {
     fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
         let mut elements = Vec::new();
         for element in &self.elements {
@@ -492,7 +480,37 @@ impl ast::ListComprehension {
     }
 }
 
-impl ast::SetComprehension {
+impl ast::ListComprehension {
+    fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
+        let values = self.value.evaluate_eager(exec)?.into_list()?;
+        let mut elements = Vec::new();
+        let mut loop_locals = VariableMap::nested(exec.locals);
+        for value in values {
+            loop_locals.clear();
+            let mut loop_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                config: exec.config,
+                locals: &mut loop_locals,
+                current_regex_captures: exec.current_regex_captures,
+                mat: exec.mat,
+                store: exec.store,
+                scoped_store: exec.scoped_store,
+                lazy_graph: exec.lazy_graph,
+                function_parameters: exec.function_parameters,
+                prev_element_debug_info: exec.prev_element_debug_info,
+                shorthands: exec.shorthands,
+            };
+            self.variable
+                .add_lazy(&mut loop_exec, value.into(), false)?;
+            let element = self.element.evaluate_lazy(&mut loop_exec)?;
+            elements.push(element);
+        }
+        Ok(elements.into())
+    }
+}
+
+impl ast::SetLiteral {
     fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
         let mut elements = Vec::new();
         for element in &self.elements {
@@ -502,6 +520,36 @@ impl ast::SetComprehension {
     }
 }
 
+impl ast::SetComprehension {
+    fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
+        let values = self.value.evaluate_eager(exec)?.into_list()?;
+        let mut elements = Vec::new();
+        let mut loop_locals = VariableMap::nested(exec.locals);
+        for value in values {
+            loop_locals.clear();
+            let mut loop_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                config: exec.config,
+                locals: &mut loop_locals,
+                current_regex_captures: exec.current_regex_captures,
+                mat: exec.mat,
+                store: exec.store,
+                scoped_store: exec.scoped_store,
+                lazy_graph: exec.lazy_graph,
+                function_parameters: exec.function_parameters,
+                prev_element_debug_info: exec.prev_element_debug_info,
+                shorthands: exec.shorthands,
+            };
+            self.variable
+                .add_lazy(&mut loop_exec, value.into(), false)?;
+            let element = self.element.evaluate_lazy(&mut loop_exec)?;
+            elements.push(element);
+        }
+        Ok(LazySet::new(elements).into())
+    }
+}
+
 impl ast::Capture {
     fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
         Ok(query_capture_value(
@@ -608,7 +656,7 @@ impl ast::ScopedVariable {
 
 impl ast::UnscopedVariable {
     fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyValue, ExecutionError> {
-        if let Some(value) = exec.globals.get(&self.name) {
+        if let Some(value) = exec.config.globals.get(&self.name) {
             Some(value.clone().into())
         } else {
             exec.locals.get(&self.name).map(|value| value.clone())
@@ -624,7 +672,7 @@ impl ast::UnscopedVariable {
         value: LazyValue,
         mutable: bool,
     ) -> Result<(), ExecutionError> {
-        if exec.globals.get(&self.name).is_some() {
+        if exec.config.globals.get(&self.name).is_some() {
             return Err(ExecutionError::DuplicateVariable(format!(
                 " global {}",
                 self
@@ -641,7 +689,7 @@ impl ast::UnscopedVariable {
         exec: &mut ExecutionContext,
         value: LazyValue,
     ) -> Result<(), ExecutionError> {
-        if exec.globals.get(&self.name).is_some() {
+        if exec.config.globals.get(&self.name).is_some() {
             return Err(ExecutionError::CannotAssignImmutableVariable(format!(
                 " global {}",
                 self
@@ -661,9 +709,53 @@ impl ast::UnscopedVariable {
 }
 
 impl ast::Attribute {
-    fn evaluate_lazy(&self, exec: &mut ExecutionContext) -> Result<LazyAttribute, ExecutionError> {
+    fn execute_lazy<F>(
+        &self,
+        exec: &mut ExecutionContext,
+        add_attribute: &mut F,
+    ) -> Result<(), ExecutionError>
+    where
+        F: FnMut(LazyAttribute) -> (),
+    {
         let value = self.value.evaluate_lazy(exec)?;
-        let attribute = LazyAttribute::new(self.name.clone(), value);
-        Ok(attribute)
+        if let Some(shorthand) = exec.shorthands.get(&self.name) {
+            shorthand.execute_lazy(exec, add_attribute, value)
+        } else {
+            add_attribute(LazyAttribute::new(self.name.clone(), value));
+            Ok(())
+        }
+    }
+}
+
+impl ast::AttributeShorthand {
+    fn execute_lazy<F>(
+        &self,
+        exec: &mut ExecutionContext,
+        add_attribute: &mut F,
+        value: LazyValue,
+    ) -> Result<(), ExecutionError>
+    where
+        F: FnMut(LazyAttribute) -> (),
+    {
+        let mut shorthand_locals = VariableMap::new();
+        let mut shorthand_exec = ExecutionContext {
+            source: exec.source,
+            graph: exec.graph,
+            config: exec.config,
+            locals: &mut shorthand_locals,
+            current_regex_captures: exec.current_regex_captures,
+            mat: exec.mat,
+            store: exec.store,
+            scoped_store: exec.scoped_store,
+            lazy_graph: exec.lazy_graph,
+            function_parameters: exec.function_parameters,
+            prev_element_debug_info: exec.prev_element_debug_info,
+            shorthands: exec.shorthands,
+        };
+        self.variable.add_lazy(&mut shorthand_exec, value, false)?;
+        for attr in &self.attributes {
+            attr.execute_lazy(&mut shorthand_exec, add_attribute)?;
+        }
+        Ok(())
     }
 }
