@@ -34,13 +34,15 @@ mod execution;
 pub mod functions;
 pub mod graph;
 mod lazy_execution;
+pub mod parse_error;
 mod parser;
 mod variables;
 
+pub use execution::ExecutionConfig;
 pub use execution::ExecutionError;
-pub use execution::Globals as Variables;
 pub use parser::Location;
 pub use parser::ParseError;
+pub use variables::Globals as Variables;
 
 use std::borrow::Borrow;
 use std::hash::Hash;
