@@ -11,7 +11,12 @@ use std::str::Chars;
 
 use regex::Regex;
 use thiserror::Error;
+use tree_sitter::CaptureQuantifier;
+use tree_sitter::CaptureQuantifier::One;
+use tree_sitter::CaptureQuantifier::OneOrMore;
 use tree_sitter::CaptureQuantifier::Zero;
+use tree_sitter::CaptureQuantifier::ZeroOrMore;
+use tree_sitter::CaptureQuantifier::ZeroOrOne;
 use tree_sitter::Language;
 use tree_sitter::Query;
 use tree_sitter::QueryError;
@@ -43,6 +48,8 @@ impl ast::File {
 /// An error that can occur while parsing a graph DSL file
 #[derive(Debug, Error)]
 pub enum ParseError {
+    #[error("Expected quantifier at {0}")]
+    ExpectedQuantifier(Location),
     #[error("Expected '{0}' at {1}")]
     ExpectedToken(&'static str, Location),
     #[error("Expected variable name at {0}")]
@@ -206,14 +213,72 @@ impl<'a> Parser<'a> {
     fn parse_into_file(&mut self, file: &mut ast::File) -> Result<(), ParseError> {
         self.consume_whitespace();
         while self.try_peek().is_some() {
-            let stanza = self.parse_stanza(file.language)?;
-            file.stanzas.push(stanza);
+            if let Ok(_) = self.consume_token("global") {
+                self.consume_whitespace();
+                let global = self.parse_global()?;
+                file.globals.push(global);
+            } else if let Ok(_) = self.consume_token("attribute") {
+                self.consume_whitespace();
+                let shorthand = self.parse_shorthand()?;
+                file.shorthands.add(shorthand);
+            } else {
+                let stanza = self.parse_stanza(file.language)?;
+                file.stanzas.push(stanza);
+            }
             self.consume_whitespace();
         }
-        file.query = Some(Query::new(file.language, &self.query_source)?);
+        // we can unwrap here because all queries have already been parsed before
+        file.query = Some(Query::new(file.language, &self.query_source).unwrap());
         Ok(())
     }
 
+    fn parse_global(&mut self) -> Result<ast::Global, ParseError> {
+        let location = self.location;
+        let name = self.parse_identifier("global variable")?;
+        let quantifier = self.parse_quantifier()?;
+        Ok(ast::Global {
+            name,
+            quantifier,
+            location,
+        })
+    }
+
+    fn parse_shorthand(&mut self) -> Result<ast::AttributeShorthand, ParseError> {
+        let location = self.location;
+        let name = self.parse_identifier("shorthand name")?;
+        self.consume_whitespace();
+        self.consume_token("=")?;
+        self.consume_whitespace();
+        let variable = self.parse_unscoped_variable()?;
+        self.consume_whitespace();
+        self.consume_token("=>")?;
+        self.consume_whitespace();
+        let attributes = self.parse_attributes()?;
+        Ok(ast::AttributeShorthand {
+            name,
+            variable,
+            attributes,
+            location,
+        })
+    }
+
+    fn parse_quantifier(&mut self) -> Result<CaptureQuantifier, ParseError> {
+        let mut quantifier = One;
+        if let Some(c) = self.try_peek() {
+            self.skip().unwrap();
+            if c == '?' {
+                quantifier = ZeroOrOne;
+            } else if c == '*' {
+                quantifier = ZeroOrMore;
+            } else if c == '+' {
+                quantifier = OneOrMore;
+            } else if !c.is_whitespace() {
+                return Err(ParseError::ExpectedQuantifier(self.location));
+            }
+        }
+        Ok(quantifier)
+    }
+
     fn parse_stanza(&mut self, language: Language) -> Result<ast::Stanza, ParseError> {
         let location = self.location;
         let query = self.parse_query(language)?;
@@ -228,16 +293,21 @@ impl<'a> Parser<'a> {
     }
 
     fn parse_query(&mut self, language: Language) -> Result<Query, ParseError> {
+        let location = self.location;
         let query_start = self.offset;
         self.skip_query()?;
         let query_end = self.offset;
         let query_source = self.source[query_start..query_end].to_owned() + "@" + FULL_MATCH;
         // If tree-sitter allowed us to incrementally add patterns to a query, we wouldn't need
-        // the global query_source and could compute the cpature indices in the AST instead of
-        // having to resolve the capture names at execution time.
+        // the global query_source.
         self.query_source += &query_source;
         self.query_source += "\n";
-        Ok(Query::new(language, &query_source)?)
+        let query = Query::new(language, &query_source).map_err(|mut e| {
+            e.row += location.row;
+            e.offset += query_start;
+            e
+        })?;
+        Ok(query)
     }
 
     fn skip_query(&mut self) -> Result<(), ParseError> {
@@ -496,12 +566,7 @@ impl<'a> Parser<'a> {
             .into())
         } else if keyword == "for" {
             self.consume_whitespace();
-            let variable = match self.parse_variable()? {
-                ast::Variable::Unscoped(variable) => Ok(variable),
-                ast::Variable::Scoped(variable) => {
-                    Err(ParseError::ExpectedUnscopedVariable(variable.location))
-                }
-            }?;
+            let variable = self.parse_unscoped_variable()?;
             self.consume_whitespace();
             self.consume_token("in")?;
             self.consume_whitespace();
@@ -666,19 +731,81 @@ impl<'a> Parser<'a> {
     }
 
     fn parse_list(&mut self) -> Result<ast::Expression, ParseError> {
+        let location = self.location;
         self.consume_token("[")?;
         self.consume_whitespace();
-        let elements = self.parse_sequence(']')?;
-        self.consume_token("]")?;
-        Ok(ast::ListComprehension { elements }.into())
+        if let Ok(_) = self.consume_token("]") {
+            return Ok(ast::ListLiteral { elements: vec![] }.into());
+        }
+        let first_element = self.parse_expression()?;
+        self.consume_whitespace();
+        if let Ok(_) = self.consume_token("]") {
+            let elements = vec![first_element];
+            Ok(ast::ListLiteral { elements }.into())
+        } else if let Ok(_) = self.consume_token(",") {
+            self.consume_whitespace();
+            let mut elements = self.parse_sequence(']')?;
+            self.consume_whitespace();
+            self.consume_token("]")?;
+            elements.insert(0, first_element);
+            Ok(ast::ListLiteral { elements }.into())
+        } else {
+            self.consume_token("for")?;
+            self.consume_whitespace();
+            let variable = self.parse_unscoped_variable()?;
+            self.consume_whitespace();
+            self.consume_token("in")?;
+            self.consume_whitespace();
+            let value = self.parse_expression()?;
+            self.consume_whitespace();
+            self.consume_token("]")?;
+            Ok(ast::ListComprehension {
+                element: first_element.into(),
+                variable,
+                value: value.into(),
+                location,
+            }
+            .into())
+        }
     }
 
     fn parse_set(&mut self) -> Result<ast::Expression, ParseError> {
+        let location = self.location;
         self.consume_token("{")?;
         self.consume_whitespace();
-        let elements = self.parse_sequence('}')?;
-        self.consume_token("}")?;
-        Ok(ast::SetComprehension { elements }.into())
+        if let Ok(_) = self.consume_token("}") {
+            return Ok(ast::SetLiteral { elements: vec![] }.into());
+        }
+        let first_element = self.parse_expression()?;
+        self.consume_whitespace();
+        if let Ok(_) = self.consume_token("}") {
+            let elements = vec![first_element];
+            Ok(ast::SetLiteral { elements }.into())
+        } else if let Ok(_) = self.consume_token(",") {
+            self.consume_whitespace();
+            let mut elements = self.parse_sequence('}')?;
+            self.consume_whitespace();
+            self.consume_token("}")?;
+            elements.insert(0, first_element);
+            Ok(ast::SetLiteral { elements }.into())
+        } else {
+            self.consume_token("for")?;
+            self.consume_whitespace();
+            let variable = self.parse_unscoped_variable()?;
+            self.consume_whitespace();
+            self.consume_token("in")?;
+            self.consume_whitespace();
+            let value = self.parse_expression()?;
+            self.consume_whitespace();
+            self.consume_token("}")?;
+            Ok(ast::SetComprehension {
+                element: first_element.into(),
+                variable,
+                value: value.into(),
+                location,
+            }
+            .into())
+        }
     }
 
     fn parse_capture(&mut self) -> Result<ast::Capture, ParseError> {
@@ -778,4 +905,13 @@ impl<'a> Parser<'a> {
             _ => Err(ParseError::ExpectedVariable(expression_location)),
         }
     }
+
+    fn parse_unscoped_variable(&mut self) -> Result<ast::UnscopedVariable, ParseError> {
+        match self.parse_variable()? {
+            ast::Variable::Unscoped(variable) => Ok(variable),
+            ast::Variable::Scoped(variable) => {
+                Err(ParseError::ExpectedUnscopedVariable(variable.location))
+            }
+        }
+    }
 }
