@@ -9,6 +9,7 @@ use indoc::indoc;
 use tree_sitter::Parser;
 use tree_sitter_graph::ast::File;
 use tree_sitter_graph::functions::Functions;
+use tree_sitter_graph::ExecutionConfig;
 use tree_sitter_graph::ExecutionError;
 use tree_sitter_graph::Identifier;
 use tree_sitter_graph::Variables;
@@ -34,7 +35,8 @@ fn execute(python_source: &str, dsl_source: &str) -> Result<String, ExecutionErr
     globals
         .add(Identifier::from("filename"), "test.py".into())
         .map_err(|_| ExecutionError::DuplicateVariable("filename".into()))?;
-    let graph = file.execute(&tree, python_source, &mut functions, &mut globals)?;
+    let mut config = ExecutionConfig::new(&mut functions, &globals);
+    let graph = file.execute(&tree, python_source, &mut config)?;
     let result = graph.pretty_print().to_string();
     Ok(result)
 }
@@ -252,6 +254,8 @@ fn can_use_global_variable() {
     check_execution(
         "pass",
         indoc! {r#"
+          global filename
+
           (module) @root
           {
             node n
@@ -265,6 +269,31 @@ fn can_use_global_variable() {
     );
 }
 
+#[test]
+fn cannot_omit_global_variable() {
+    fail_execution(
+        "pass",
+        indoc! {r#"
+          global root
+
+          (identifier) {
+            node n
+            edge n -> root
+          }
+        "#},
+    );
+}
+
+#[test]
+fn cannot_pass_string_to_global_list_variable() {
+    fail_execution(
+        "pass",
+        indoc! {r#"
+          global filename*
+        "#},
+    );
+}
+
 #[test]
 fn can_use_variable_multiple_times() {
     check_execution(
@@ -768,6 +797,50 @@ fn variables_are_inherited_in_for_in_body() {
     );
 }
 
+#[test]
+fn can_execute_list_comprehension() {
+    check_execution(
+        r#"
+          pass
+          pass
+          pass
+        "#,
+        indoc! {r#"
+          (module (pass_statement)* @xs) @root
+          {
+            node node0
+            attr (node0) val = [ (named-child-index x) for x in @xs ]
+          }
+        "#},
+        indoc! {r#"
+          node 0
+            val: [0, 1, 2]
+        "#},
+    );
+}
+
+#[test]
+fn can_execute_set_comprehension() {
+    check_execution(
+        r#"
+          pass
+          pass
+          pass
+        "#,
+        indoc! {r#"
+          (module (pass_statement)* @xs) @root
+          {
+            node node0
+            attr (node0) val = { (source-text x) for x in @xs }
+          }
+        "#},
+        indoc! {r#"
+          node 0
+            val: {"pass"}
+        "#},
+    );
+}
+
 #[test]
 fn can_execute_scan_of_local_call_expression() {
     check_execution(
@@ -820,3 +893,25 @@ fn can_execute_scan_of_local_variable() {
         "#},
     );
 }
+
+#[test]
+fn can_execute_shorthand() {
+    check_execution(
+        indoc! { r#"
+          def get_f():
+            pass
+        "#},
+        indoc! {r#"
+            attribute def = x => source_node = x, symbol = (source-text x)
+            (function_definition name: (identifier) @name) {
+              node n
+              attr (n) def = @name
+            }
+        "#},
+        indoc! {r#"
+          node 0
+            source_node: [syntax node identifier (1, 5)]
+            symbol: "get_f"
+        "#},
+    );
+}
