@@ -337,7 +337,7 @@ fn can_parse_lists() {
                     location: Location { row: 3, column: 14 }
                 }
                 .into(),
-                value: ListComprehension {
+                value: ListLiteral {
                     elements: vec![
                         IntegerConstant { value: 1 }.into(),
                         IntegerConstant { value: 2 }.into(),
@@ -354,7 +354,7 @@ fn can_parse_lists() {
                     location: Location { row: 4, column: 14 }
                 }
                 .into(),
-                value: ListComprehension { elements: vec![] }.into(),
+                value: ListLiteral { elements: vec![] }.into(),
                 location: Location { row: 4, column: 10 },
             }
             .into(),
@@ -364,7 +364,7 @@ fn can_parse_lists() {
                     location: Location { row: 5, column: 14 }
                 }
                 .into(),
-                value: ListComprehension {
+                value: ListLiteral {
                     elements: vec![
                         StringConstant {
                             value: String::from("hello")
@@ -414,7 +414,7 @@ fn can_parse_sets() {
                     location: Location { row: 3, column: 14 }
                 }
                 .into(),
-                value: SetComprehension {
+                value: SetLiteral {
                     elements: vec![
                         IntegerConstant { value: 1 }.into(),
                         IntegerConstant { value: 2 }.into(),
@@ -431,7 +431,7 @@ fn can_parse_sets() {
                     location: Location { row: 4, column: 14 }
                 }
                 .into(),
-                value: SetComprehension { elements: vec![] }.into(),
+                value: SetLiteral { elements: vec![] }.into(),
                 location: Location { row: 4, column: 10 },
             }
             .into(),
@@ -441,7 +441,7 @@ fn can_parse_sets() {
                     location: Location { row: 5, column: 14 }
                 }
                 .into(),
-                value: SetComprehension {
+                value: SetLiteral {
                     elements: vec![
                         StringConstant {
                             value: String::from("hello")
@@ -1074,3 +1074,448 @@ fn cannot_parse_scan_of_nonlocal_variable() {
         panic!("Parse succeeded unexpectedly");
     }
 }
+
+#[test]
+fn can_parse_list_comprehension() {
+    let source = r#"
+        (module (_)* @xs)@mod
+        {
+          print [ (named-child-index x) for x in @xs ]
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![Print {
+            values: vec![ListComprehension {
+                element: Box::new(
+                    Call {
+                        function: "named-child-index".into(),
+                        parameters: vec![UnscopedVariable {
+                            name: "x".into(),
+                            location: Location { row: 3, column: 37 }
+                        }
+                        .into()]
+                    }
+                    .into()
+                ),
+                variable: UnscopedVariable {
+                    name: "x".into(),
+                    location: Location { row: 3, column: 44 }
+                },
+                value: Box::new(
+                    Capture {
+                        name: "xs".into(),
+                        quantifier: ZeroOrMore,
+                        file_capture_index: 0,
+                        stanza_capture_index: 0,
+                        location: Location { row: 3, column: 49 }
+                    }
+                    .into()
+                ),
+                location: Location { row: 3, column: 16 }
+            }
+            .into()],
+            location: Location { row: 3, column: 10 }
+        }
+        .into()]]
+    );
+}
+
+#[test]
+fn can_parse_set_comprehension() {
+    let source = r#"
+        (module (_)* @xs)@mod
+        {
+          print { (named-child-index x) for x in @xs }
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![Print {
+            values: vec![SetComprehension {
+                element: Box::new(
+                    Call {
+                        function: "named-child-index".into(),
+                        parameters: vec![UnscopedVariable {
+                            name: "x".into(),
+                            location: Location { row: 3, column: 37 }
+                        }
+                        .into()]
+                    }
+                    .into()
+                ),
+                variable: UnscopedVariable {
+                    name: "x".into(),
+                    location: Location { row: 3, column: 44 }
+                },
+                value: Box::new(
+                    Capture {
+                        name: "xs".into(),
+                        quantifier: ZeroOrMore,
+                        file_capture_index: 0,
+                        stanza_capture_index: 0,
+                        location: Location { row: 3, column: 49 }
+                    }
+                    .into()
+                ),
+                location: Location { row: 3, column: 16 }
+            }
+            .into()],
+            location: Location { row: 3, column: 10 }
+        }
+        .into()]]
+    );
+}
+
+#[test]
+fn can_parse_global() {
+    let source = r#"
+        global root
+
+        (identifier) {
+          node n
+          edge n -> root
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    assert_eq!(
+        file.globals,
+        vec![Global {
+            name: "root".into(),
+            quantifier: One,
+            location: Location { row: 1, column: 15 },
+        }]
+    );
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![
+            CreateGraphNode {
+                node: UnscopedVariable {
+                    name: "n".into(),
+                    location: Location { row: 4, column: 15 },
+                }
+                .into(),
+                location: Location { row: 4, column: 10 },
+            }
+            .into(),
+            CreateEdge {
+                source: UnscopedVariable {
+                    name: "n".into(),
+                    location: Location { row: 5, column: 15 },
+                }
+                .into(),
+                sink: UnscopedVariable {
+                    name: "root".into(),
+                    location: Location { row: 5, column: 20 },
+                }
+                .into(),
+                location: Location { row: 5, column: 10 },
+            }
+            .into(),
+        ]]
+    );
+}
+
+#[test]
+// TODO deprecated support for undeclared globals
+//      change to cannot_parse_undeclared_global once support is removed
+fn can_parse_undeclared_global() {
+    let source = r#"
+        (identifier) {
+          node n
+          edge n -> root
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    assert_eq!(file.globals, vec![]);
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![
+            CreateGraphNode {
+                node: UnscopedVariable {
+                    name: "n".into(),
+                    location: Location { row: 2, column: 15 },
+                }
+                .into(),
+                location: Location { row: 2, column: 10 },
+            }
+            .into(),
+            CreateEdge {
+                source: UnscopedVariable {
+                    name: "n".into(),
+                    location: Location { row: 3, column: 15 },
+                }
+                .into(),
+                sink: UnscopedVariable {
+                    name: "root".into(),
+                    location: Location { row: 3, column: 20 },
+                }
+                .into(),
+                location: Location { row: 3, column: 10 },
+            }
+            .into(),
+        ]]
+    );
+}
+
+#[test]
+fn can_parse_list_global() {
+    let source = r#"
+        global roots*
+
+        (identifier) {
+          for root in roots {
+            node n
+            edge n -> root
+          }
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    assert_eq!(
+        file.globals,
+        vec![Global {
+            name: "roots".into(),
+            quantifier: ZeroOrMore,
+            location: Location { row: 1, column: 15 },
+        }]
+    );
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![ForIn {
+            variable: UnscopedVariable {
+                name: "root".into(),
+                location: Location { row: 4, column: 14 },
+            }
+            .into(),
+            value: UnscopedVariable {
+                name: "roots".into(),
+                location: Location { row: 4, column: 22 },
+            }
+            .into(),
+            statements: vec![
+                CreateGraphNode {
+                    node: UnscopedVariable {
+                        name: "n".into(),
+                        location: Location { row: 5, column: 17 },
+                    }
+                    .into(),
+                    location: Location { row: 5, column: 12 },
+                }
+                .into(),
+                CreateEdge {
+                    source: UnscopedVariable {
+                        name: "n".into(),
+                        location: Location { row: 6, column: 17 },
+                    }
+                    .into(),
+                    sink: UnscopedVariable {
+                        name: "root".into(),
+                        location: Location { row: 6, column: 22 },
+                    }
+                    .into(),
+                    location: Location { row: 6, column: 12 },
+                }
+                .into(),
+            ],
+            location: Location { row: 4, column: 10 },
+        }
+        .into(),]]
+    );
+}
+
+#[test]
+fn can_parse_optional_global() {
+    let source = r#"
+        global root?
+
+        (identifier) {
+          if some root {
+            node n
+            edge n -> root
+          }
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    assert_eq!(
+        file.globals,
+        vec![Global {
+            name: "root".into(),
+            quantifier: ZeroOrOne,
+            location: Location { row: 1, column: 15 },
+        }]
+    );
+
+    let statements = file
+        .stanzas
+        .into_iter()
+        .map(|s| s.statements)
+        .collect::<Vec<_>>();
+    assert_eq!(
+        statements,
+        vec![vec![If {
+            arms: vec![IfArm {
+                conditions: vec![Condition::Some {
+                    value: UnscopedVariable {
+                        name: "root".into(),
+                        location: Location { row: 4, column: 18 },
+                    }
+                    .into(),
+                    location: Location { row: 4, column: 13 },
+                },],
+                statements: vec![
+                    CreateGraphNode {
+                        node: UnscopedVariable {
+                            name: "n".into(),
+                            location: Location { row: 5, column: 17 },
+                        }
+                        .into(),
+                        location: Location { row: 5, column: 12 },
+                    }
+                    .into(),
+                    CreateEdge {
+                        source: UnscopedVariable {
+                            name: "n".into(),
+                            location: Location { row: 6, column: 17 },
+                        }
+                        .into(),
+                        sink: UnscopedVariable {
+                            name: "root".into(),
+                            location: Location { row: 6, column: 22 },
+                        }
+                        .into(),
+                        location: Location { row: 6, column: 12 },
+                    }
+                    .into(),
+                ],
+                location: Location { row: 4, column: 10 },
+            },],
+            location: Location { row: 4, column: 10 },
+        }
+        .into(),]]
+    );
+}
+
+#[test]
+fn cannot_parse_global_with_unknown_quantifier() {
+    let source = r#"
+        global root%
+
+        (module) {
+          node root
+        }
+    "#;
+    if let Ok(_) = File::from_str(tree_sitter_python::language(), source) {
+        panic!("Parse succeeded unexpectedly");
+    }
+}
+
+#[test]
+fn cannot_parse_hiding_global() {
+    let source = r#"
+        global root
+
+        (module) {
+          node root
+        }
+    "#;
+    if let Ok(_) = File::from_str(tree_sitter_python::language(), source) {
+        panic!("Parse succeeded unexpectedly");
+    }
+}
+
+#[test]
+fn cannot_parse_set_global() {
+    let source = r#"
+        global root
+
+        (module) {
+          set root = #null
+        }
+    "#;
+    if let Ok(_) = File::from_str(tree_sitter_python::language(), source) {
+        panic!("Parse succeeded unexpectedly");
+    }
+}
+
+#[test]
+fn can_parse_shorthand() {
+    let source = r#"
+        attribute def = x => source_node = x, symbol = (source-text x)
+        (function_definition name: (identifier) @name) {
+          node n
+          attr (n) sh = @name
+        }
+    "#;
+    let file = File::from_str(tree_sitter_python::language(), source).expect("Cannot parse file");
+
+    let shorthands = file.shorthands.into_iter().collect::<Vec<_>>();
+    assert_eq!(
+        shorthands,
+        vec![AttributeShorthand {
+            name: "def".into(),
+            variable: UnscopedVariable {
+                name: "x".into(),
+                location: Location { row: 1, column: 24 }
+            },
+            attributes: vec![
+                Attribute {
+                    name: "source_node".into(),
+                    value: UnscopedVariable {
+                        name: "x".into(),
+                        location: Location { row: 1, column: 43 }
+                    }
+                    .into()
+                },
+                Attribute {
+                    name: "symbol".into(),
+                    value: Call {
+                        function: "source-text".into(),
+                        parameters: vec![UnscopedVariable {
+                            name: "x".into(),
+                            location: Location { row: 1, column: 68 }
+                        }
+                        .into()]
+                    }
+                    .into(),
+                }
+            ],
+            location: Location { row: 1, column: 18 }
+        }]
+    );
+}
